/* Simple program:  Move N sprites around on the screen as fast as possible */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <time.h>

#include "SDL.h"

#define DEBUG_FLIP 1

#define NUM_SPRITES	49
#define MAX_SPEED 	1

SDL_Surface *sprite;
int numsprites;
SDL_Rect *sprite_rects;
SDL_Rect *positions;
SDL_Rect *velocities;
Uint16 sprite_w, sprite_h;
#define false 0
#define true 1
#define time_elapsed 0.033f //pretend that much time elaped each iteration

#pragma comment(lib, "SDL.lib")
#pragma comment(lib, "SDLmain.lib")


int   bounce_x = 1;
int   bounce_y = 1;
int   down = 0;
int   right = 0;

typedef struct nodes_s
{
	float x[2], y[2], vx, vy;
	struct nodes_s *neigbor[8];
	float ne[8];
}node_t;
node_t nodes[49];

void putpixel(SDL_Surface *surface, int x, int y)
{
    int bpp = surface->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to set */
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;
    int pixel = 0xFFFFFFFF;

    switch(bpp) {
    case 1:
        *p = pixel;
        break;

    case 2:
        *(Uint16 *)p = pixel;
        break;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
            p[0] = (pixel >> 16) & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = pixel & 0xff;
        } else {
            p[0] = pixel & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = (pixel >> 16) & 0xff;
        }
        break;

    case 4:
        *(Uint32 *)p = pixel;
        break;
    }
}

drawline(SDL_Surface *screen, int x, int y, int x2, int y2)
{
  int j;
  int yL = false;
  int sL = y2-y;
  int lL = x2-x;
  if(abs(sL)>abs(lL)){
    int swap = sL;
    sL = lL;
    lL = swap;
    yL = true;
  }
  int dI;
  if(lL == 0) dI = 0;
  else dI = (sL << 16)/lL;
  if(yL){
    if(lL>0){
      lL += y;
      for(j=0x8000+(x << 16); y<=lL; ++y){
        putpixel(screen ,j >> 16, y); //define your own putPix mofo
	j += dI;
      }
      return;
    }
    lL += y;
    for(j=0x8000+(x << 16); y>=lL; --y){
      putpixel(screen, j >> 16, y);
      j -= dI;
    }
    return;
  }
  if(lL>0){
    lL += x;
    for(j=0x8000+(y << 16); x<=lL; ++x){
      putpixel(screen, x, j >> 16);
      j += dI;
    }
    return;
  }
  lL += x;
  for(j=0x8000+(y << 16); x>=lL; --x){
    putpixel(screen, x, j >> 16);
    j -= dI;
  }
}

int LoadSprite(SDL_Surface *screen, char *file)
{
	SDL_Surface *temp;
	

	/* Load the sprite image */
	sprite = SDL_LoadBMP(file);
	if ( sprite == NULL ) {
		fprintf(stderr, "Couldn't load %s: %s", file, SDL_GetError());
		return(-1);
	}

	/* Set transparent pixel as the pixel at (0,0) */
	if ( sprite->format->palette ) {
		SDL_SetColorKey(sprite, (SDL_SRCCOLORKEY|SDL_RLEACCEL),
						*(Uint8 *)sprite->pixels);
	}

	/* Convert sprite to video format */
	temp = SDL_DisplayFormat(sprite);
	SDL_FreeSurface(sprite);
	if ( temp == NULL ) {
		fprintf(stderr, "Couldn't convert background: %s\n",
							SDL_GetError());
		return(-1);
	}
	sprite = temp;

	/* We're ready to roll. :) */
	return(0);
}

void MoveSprites(SDL_Surface *screen, Uint32 background)
{
	static int old = 0;
	static int newx = 1;
	float force;	
	float dx, dy, d;
	int l;
	int sumx = 0;
	int sumy = 0;
	int i, nupdates;
	SDL_Rect area;
	SDL_Rect *dstrect = NULL;

	nupdates = 0;
	
	
        SDL_FillRect(screen, dstrect, 0);


	/* Move the sprite, bounce at the wall, and draw */
	for ( i=0; i<numsprites; ++i ) {
		for(l = 0; l<8 && nodes[i].neigbor[l] != NULL;l++) {
			dx = nodes[i].neigbor[l]->x[old] - nodes[i].x[old];
			dy = nodes[i].neigbor[l]->y[old] - nodes[i].y[old];

			if((dx == 0.0f)&&(dy == 0.0f))
				d = 0;
			else if (dx == 0.0f) {
				d = fabsf(dy);
				dy=dy/d;
			}
			else if (dy == 0.0f) {
                                d = fabsf(dx);
                                dx=dx/d;
                        }
			else {
			d = sqrt(float(dx*dx)+(dy*dy)); //pythagorean distance
			dx /= d;
			dy /= d;
			}
			d -= nodes[i].ne[l]; //get distance relative to equalibrium
			force = (200.0*d);
         if ( force > 0 )
            force -= 10;
         //   printf( "positive\n" );
         else
            force += 10;
         //   printf( "negative\n" );
			sumx += dx*force; //cos*k*d
			sumy += dy*force; //sin*k*d
		}
		nodes[i].vx += (sumx*time_elapsed) + right * 0.1;
		nodes[i].vy += (sumy*time_elapsed) + down * 0.1;
		nodes[i].x[newx] = nodes[i].x[old] + (nodes[i].vx*time_elapsed);
		nodes[i].y[newx] = nodes[i].y[old] + (nodes[i].vy*time_elapsed);

		if (nodes[i].x[newx] < sprite_w/2) {
			nodes[i].x[newx] = sprite_w/2;
         nodes[i].vx *= 0.0;
		//	nodes[i].vx = -nodes[i].vx;
		}
		if (nodes[i].x[newx] > (screen->w - sprite_w/2)) {
			nodes[i].x[newx] = screen->w - sprite_w/2;
         nodes[i].vx *= 0.0;
		//	nodes[i].vx = -nodes[i].vx;
		}
		if (nodes[i].y[newx] < sprite_w/2) {
			nodes[i].y[newx] = sprite_w/2;
         nodes[i].vy *= 0.0;
		//	nodes[i].vy = -nodes[i].vy;
		}
		if (nodes[i].y[newx] > (screen->h - sprite_w/2)) {
			nodes[i].y[newx] = screen->h - sprite_w/2;
         nodes[i].vy *= -0.0;
         //	nodes[i].vy = -nodes[i].vy;
		}

		sumx = 0;
		sumy = 0;
	}
	for ( i=0; i<numsprites; i+=2 ) {

		for(l = 0; l<8 && nodes[i].neigbor[l] != NULL; l++) {
			drawline(screen, (int)nodes[i].neigbor[l]->x[newx], (int)nodes[i].neigbor[l]->y[newx], (int)nodes[i].x[newx], (int)nodes[i].y[newx]);
		}
	}
	for ( i=0; i<numsprites; ++i ) {


		area.x = (int)nodes[i].x[newx] - sprite_w/2;
		area.y = (int)nodes[i].y[newx] - sprite_w/2;
		
		SDL_BlitSurface(sprite, NULL, screen, &area);
		sprite_rects[nupdates++] = area;
	}
	i = old;
	old = newx;
	newx = i;
        SDL_UpdateRect(screen, 0, 0, 0, 0);

	/* Update the screen! */
	if ( (screen->flags & SDL_DOUBLEBUF) == SDL_DOUBLEBUF ) {
		SDL_Flip(screen);
	} else {
		SDL_UpdateRects(screen, nupdates, sprite_rects);
	}

}

/* This is a way of telling whether or not to use hardware surfaces */
Uint32 FastestFlags(Uint32 flags, int width, int height, int bpp)
{
	const SDL_VideoInfo *info;

	/* Hardware acceleration is only used in fullscreen mode */
	flags |= SDL_FULLSCREEN;

	/* Check for various video capabilities */
	info = SDL_GetVideoInfo();
	if ( info->blit_hw_CC && info->blit_fill ) {
		/* We use accelerated colorkeying and color filling */
		flags |= SDL_HWSURFACE;
	}
	/* If we have enough video memory, and will use accelerated
	   blits directly to it, then use page flipping.
	 */
	if ( (flags & SDL_HWSURFACE) == SDL_HWSURFACE ) {
		/* Direct hardware blitting without double-buffering
		   causes really bad flickering.
		 */
		if ( info->video_mem*1024 > (height*width*bpp/8) ) {
			flags |= SDL_DOUBLEBUF;
		} else {
			flags &= ~SDL_HWSURFACE;
		}
	}

	/* Return the flags */
	return(flags);
}

int main(int argc, char *argv[])
{
	SDL_Surface *screen;
	Uint8 *mem;
	int width, height;
	Uint8  video_bpp;
	Uint32 videoflags;
	Uint32 background;
	int    i, done;
	SDL_Event event;
	Uint32 then, now, frames;

	/* Initialize SDL */
	if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "Couldn't initialize SDL: %s\n",SDL_GetError());
		exit(1);
	}
	atexit(SDL_Quit);

	numsprites = NUM_SPRITES;
	videoflags = SDL_SWSURFACE|SDL_ANYFORMAT;
	width = 640;
	height = 480;
	video_bpp = 8;
	while ( argc > 1 ) {
		--argc;
		if ( strcmp(argv[argc-1], "-width") == 0 ) {
			width = atoi(argv[argc]);
			--argc;
		} else
		if ( strcmp(argv[argc-1], "-height") == 0 ) {
			height = atoi(argv[argc]);
			--argc;
		} else
		if ( strcmp(argv[argc-1], "-bpp") == 0 ) {
			video_bpp = atoi(argv[argc]);
			videoflags &= ~SDL_ANYFORMAT;
			--argc;
		} else
		if ( strcmp(argv[argc], "-fast") == 0 ) {
			videoflags = FastestFlags(videoflags, width, height, video_bpp);
		} else
		if ( strcmp(argv[argc], "-hw") == 0 ) {
			videoflags ^= SDL_HWSURFACE;
		} else
		if ( strcmp(argv[argc], "-flip") == 0 ) {
			videoflags ^= SDL_DOUBLEBUF;
		} else
		if ( strcmp(argv[argc], "-fullscreen") == 0 ) {
			videoflags ^= SDL_FULLSCREEN;
		} else
		if ( isdigit(argv[argc][0]) ) {
			numsprites = atoi(argv[argc]);
		} else {
			fprintf(stderr, 
	"Usage: %s [-bpp N] [-hw] [-flip] [-fast] [-fullscreen] [numsprites]\n",
								argv[0]);
			exit(1);
		}
	}

	/* Set video mode */
	screen = SDL_SetVideoMode(width, height, video_bpp, videoflags);
	if ( ! screen ) {
		fprintf(stderr, "Couldn't set %dx%d video mode: %s\n",
					width, height, SDL_GetError());
		exit(2);
	}

	/* Load the sprite */
	if ( LoadSprite(screen, "icon.bmp") < 0 ) {
		exit(1);
	}

	/* Allocate memory for the sprite info */
	mem = (Uint8 *)malloc(4*sizeof(SDL_Rect)*numsprites);
	if ( mem == NULL ) {
		SDL_FreeSurface(sprite);
		fprintf(stderr, "Out of memory!\n");
		exit(2);
	}
	sprite_rects = (SDL_Rect *)mem;
	positions = sprite_rects;
	sprite_rects += numsprites;
	velocities = sprite_rects;
	sprite_rects += numsprites;
	sprite_w = sprite->w;
	sprite_h = sprite->h;
	srand(time(NULL));
	
	int row =0, col =0;
	i = 0;
	int n = 0;
	for(row = 0; row<7; row++) {
		for(col = 0; col<7; col++) {
			nodes[i].x[0] = (col*40)+ .5*(width - (7*40));
			nodes[i].y[0] = (row*40)+ .5*(height - (7*40));
			nodes[i].x[1] = (col*40)+ .5*(width - (7*40));
                        nodes[i].y[1] = (row*40)+ .5*(height - (7*40));
			nodes[i].vx = 0.0f;
			nodes[i].vy = 20;
			
			if(col != 0) {
				nodes[i].neigbor[n] = &nodes[row*7 + col-1]; //left
				nodes[i].ne[n] = 40.0f;
				n++;
			}
			if((col != 0) && (row !=0)) {
				nodes[i].neigbor[n] = &nodes[(row-1)*7 + col-1]; //up left
                               nodes[i].ne[n] = 56.568542;
				n++;
			}
			if(row !=0) {
				nodes[i].neigbor[n] = &nodes[(row-1)*7 + col]; //up
                                nodes[i].ne[n] = 40.0f;
				n++;
			}
			if((col < 6) && (row !=0)) {
				nodes[i].neigbor[n] = &nodes[(row-1)*7 + col+1]; //up right
				nodes[i].ne[n] = 56.568542;
				n++;
			}
			if(col < 6) {
				nodes[i].neigbor[n] = &nodes[(row)*7 + col+1]; //right
				nodes[i].ne[n] = 40.0f;
				n++;
			}
			if((col < 6) && (row <6)) {
				nodes[i].neigbor[n] = &nodes[(row+1)*7 + col+1]; //down right
				nodes[i].ne[n] = 56.568542;
				n++;
			}
			if(row < 6) {
				nodes[i].neigbor[n] = &nodes[(row+1)*7 + col]; //down
				nodes[i].ne[n] = 40.0f;
				n++;
			}
			if((row < 6) && (col != 0)) {
				nodes[i].neigbor[n] = &nodes[(row+1)*7 + col-1]; //down left
				nodes[i].ne[n] = 56.568542;
				n++;
			}
			if(n<8) {
				nodes[i].neigbor[n] = NULL;
			}
			n = 0;
			i++;
		}
	}
	
	background = SDL_MapRGB(screen->format, 0x00, 0x00, 0x00);

	/* Print out information about our surfaces */
	printf("Screen is at %d bits per pixel\n",screen->format->BitsPerPixel);
	if ( (screen->flags & SDL_HWSURFACE) == SDL_HWSURFACE ) {
		printf("Screen is in video memory\n");
	} else {
		printf("Screen is in system memory\n");
	}
	if ( (screen->flags & SDL_DOUBLEBUF) == SDL_DOUBLEBUF ) {
		printf("Screen has double-buffering enabled\n");
	}
	if ( (sprite->flags & SDL_HWSURFACE) == SDL_HWSURFACE ) {
		printf("Sprite is in video memory\n");
	} else {
		printf("Sprite is in system memory\n");
	}
	/* Run a sample blit to trigger blit acceleration */
	{ SDL_Rect dst;
		dst.x = 0;
		dst.y = 0;
		dst.w = sprite->w;
		dst.h = sprite->h;
		SDL_BlitSurface(sprite, NULL, screen, &dst);
		SDL_FillRect(screen, &dst, background);
	}
	if ( (sprite->flags & SDL_HWACCEL) == SDL_HWACCEL ) {
		printf("Sprite blit uses hardware acceleration\n");
	}
	if ( (sprite->flags & SDL_RLEACCEL) == SDL_RLEACCEL ) {
		printf("Sprite blit uses RLE acceleration\n");
	}

	/* Loop, blitting sprites and waiting for a keystroke */
	frames = 0;
	then = SDL_GetTicks();
	done = 0;
	while ( !done ) {
		/* Check for events */
		++frames;
		while ( SDL_PollEvent(&event) ) {
			switch (event.type) {
				case SDL_MOUSEBUTTONDOWN:
					SDL_WarpMouse(screen->w/2, screen->h/2);
					break;
				case SDL_KEYDOWN:
					/* Any keypress quits the app... */
               {
                  switch (event.key.keysym.sym)
                  {
                  case SDLK_UP:
                     --down;
                     break;
                  case SDLK_DOWN:
                     ++down;
                     break;
                  case SDLK_LEFT:
                     --right;
                     break;
                  case SDLK_RIGHT:
                     ++right;
                     break;
                  }
               }
               
               break;
            //case SDL_KEYUP:
            //      up++;
            //      break;
            //case SDL_KEY
				case SDL_QUIT:
					done = 1;
					break;
				default:
					break;
			}
		}
		/* Slow the frame rate down for super fast computers */
		if(frames=50) {
			now = SDL_GetTicks();
                	if ((now-then) < 833)
				now = 1;
			else 
				now = 0;
		}
		if (now) SDL_Delay(15);
      		

		MoveSprites(screen, background);
	}
	SDL_FreeSurface(sprite);
	free(mem);

	/* Print out some timing information */
	now = SDL_GetTicks();
	if ( now > then ) {
		printf("%2.2f frames per second\n",
					((double)frames*1000)/(now-then));
	}
	SDL_Quit();
	return(0);
}
